part of disposable;

/// Provides a method of encapsulating a series of actions which ensure proper disposal of
/// the disposable instance prior to exiting but following the completion of a Future value
/// which is expected to be emitted from the contained code block
Future usingAsync(IDisposable disposable, dynamic UsingFunction(IDisposable disposable)) {
  // If the disposable was already disposed then return null
  if(disposable.isDisposed) { return new Future.value(null); }
  // Placeholder for result
  dynamic result;
  dynamic error;
  // Check for async disposable
  if(disposable is! IAsyncDisposable) { // Is an IDisposable
    Completer completer = new Completer();
    try { // Try to apply the function to the disposable
      result = UsingFunction(disposable);
    } catch(ex) {
      error = ex;
    }
    try {
      disposable.dispose();
    } catch(ex) {
      error = ex;
    }
    if(error != null) {
      completer.completeError(error);
    } else if (result == null) {
      completer.complete();
    } else {
      completer.complete(result);
    }
    return completer.future;
  } else { // Is an IAsyncDisposable
    // Get a reference to the provided future
    Future future = disposable.future;
    try { // Try to apply the function to the disposable
      result = UsingFunction(disposable);
    } catch(ex) { // Error in using function
      _tryDisposeError(disposable, ex);
      // Something has gone wrong at this point so bailing
      return future;
    }

    // Dispose the target
    if(result == null) {
      // No value to dispose with
      _tryDispose(disposable);
    } else if(result is Future) {
      // Do something later
      result.then((value) {
        // Got value from future
        _tryDisposeWith(disposable, value);
      }, onError: (error) {
        _tryDisposeError(disposable, error);
      });
    } else {
      // A value that isn't a future
      _tryDisposeWith(disposable, result);
    }
    // Return future from AsyncDisposable
    return future;
  }
  return new Future.value(null);
}

void _tryDispose(IDisposable disposable) {
  try {
    disposable.dispose();
  } catch(ex) {
    if(disposable is IAsyncDisposable) {
      disposable.disposeError(ex);
    }
  }
}

void _tryDisposeWith(IAsyncDisposable disposable, dynamic value) {
  try {
    disposable.disposeWith(value);
  } catch(ex) {
    disposable.disposeError(ex);
  }
}

void _tryDisposeError(IAsyncDisposable disposable, dynamic error) {
  try {
    disposable.disposeError(error);
  } catch(ex) {
    disposable.disposeError(ex);
  }
}
