part of disposable_test;

class Thing implements IAsyncDisposable {
  final AsyncDisposable _disposable;
  bool _isDisposed = false;

  bool get isDisposed => _isDisposed;
  Future get future => _disposable.future;

  Thing._(AsyncDisposable this._disposable);

  factory Thing() {
    Thing thing;
    AsyncDisposable disposable = new AsyncDisposable(() {
    });
    thing = new Thing._(disposable);
    return thing;
  }

  void _dispose() {
    _isDisposed = true;
  }

  void dispose() {
    _dispose();
    _disposable.dispose();
  }

  void disposeWith(dynamic value) {
    _dispose();
    _disposable.disposeWith(value);
  }

  void disposeError(dynamic error) {
    _dispose();
    _disposable.disposeError(error);
  }
}

using_async_tests() {

  group('-using_async- should', () {

    test('dispose of Disposable', () {
      // Arrange
      bool disposed = false;
      IDisposable disposable = new Disposable(() => disposed = true);

      // Act
      usingAsync(disposable, (IDisposable disp) {
        expect(disposed, false);
      })
      .then((value) {
        expect(disposed, true);
      })
      .then(expectAsync1((_) { }));
    });

    test('still trigger dispose if wrapped block throws exception', () {
      // Arrange
      bool disposed = false;
      Disposable disposable = new Disposable(() => disposed = true);
      // Act
      usingAsync(disposable, (IDisposable disposable) {
        throw 'error';
      })
      .then((value) { },
          onError: (error) { })
      .then(expectAsync1((_) { }));

      // Assert
      expect(disposed, true, reason: 'still trigger dispose if wrapped block throws exception');
    });

    test('rethrow caught exception after disposing', () {
      // Arrange
      bool disposed = false;
      bool caughtError = false;
      var error = null;
      Disposable disposable = new Disposable(() => disposed = true);

      // Act
      usingAsync(disposable, (IDisposable disposable) {
        throw 'error';
      })
      .then((value) { }, onError: (error) {
        // Assert
        expect(disposed, true);
        expect(error, 'error');
      });
    });

    test('trigger Future.completeError if disposable is IAsyncDisposable and exception is thrown in wrapped block', () {
      // Arrange
      bool disposed = false;
      bool futureCalled = false;
      var error = null;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);

      var future = disposable.future
          .then((_) => futureCalled = true)
          .catchError((ex) => error = ex)
          .then(expectAsync1((_) {
            expect(futureCalled, false);
            expect(error, 'error', reason: 'if the future error handler was triggered this will not be null');
          }));

      // Act
      try {
        using(disposable, (IAsyncDisposable disposable) {
          throw 'error';
        });
      } catch(ex) { }

      expect(disposed, true);
    });

    test('dispose of AsyncDisposable implementation with future return value', () {
      // Arrange
      Thing thing = new Thing();

      expect(thing.isDisposed, false);
      // Act
      usingAsync(thing, (Thing thing) {
        expect(thing.isDisposed, false);
      })
      .then((value) {
        expect(thing.isDisposed, true);
      })
      .then(expectAsync1((_) { }));
    });

    test('return future<null> if disposable was already disposed', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);
      disposable.dispose();

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        // Already disposed
        expect(false, true, reason: 'should not get here');
      })
      .then((value) {
        // Assert
        expect(value, null);
      })
      .then(expectAsync1((_) { }));
    });

    test('allow nothing response from wrapped code without blocking', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        expect(disposed, false, reason: 'should not have disposed inside the wrapped');
        // nothing...
      })
      .then((value) {
        // Assert
        expect(disposed, true, reason: 'should have disposed prior to completion without blocking');
      }, onError: (error) {
        expect(false, true, reason: 'should not complete with an exception: ${error}');
      })
      .then(expectAsync1((_) { }));
    });

    test('complete with result if Future is returned', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);
      Completer externalCompleter = new Completer();

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        return externalCompleter.future;
      })
      .then((value) {
        expect(disposed, true);
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true, reason: 'should not complete with an exception: ${error}');
      })
      .then(expectAsync1((_) { }));

      externalCompleter.complete('stuff');
    });

    /* ** Completer is blocked waiting for dispose
    test('complete with result if Future is completed inline', () {
      print('complete with result if Future is completed inline');
      // Arrange
      bool disposed = false;
      Completer externalCompleter = new Completer();
      IAsyncDisposable disposable = new AsyncDisposable(() {
        disposed = true;
        externalCompleter.complete('stuff');
      });

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        return externalCompleter.future;
      })
      .then((value) {
        expect(disposed, true);
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true, reason: 'should not complete with an exception: ${error}');
      })
      .then(expectAsync1((_) { }));
    });*/

    test('complete with result if Future is returned', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);
      Completer externalCompleter = new Completer();

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        return 'stuff';
      })
      .then((value) {
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true, reason: 'should not complete with an exception: ${error}');
      })
      .then(expectAsync1((_) { }));
    });

    test('complete with onError from Future if exception is thrown in wrapped code', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => disposed = true);

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false, reason: 'should not have disposed inside the wrapped');
        throw 'error';
      })
      .then((value) {
        expect(false, true, reason: 'should complete with an exception if wrapped throws');
      }, onError: (error) {
        // Assert
        expect(error, 'error');
        expect(disposed, true, reason: 'should have disposed prior to completion with error: ${error}');
      })
      .then(expectAsync1((_) { }));
    });

    test('complete with onError from Future if exception is thrown in wrapped code and dispose', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => throw 'dispose error');

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false, reason: 'should not have disposed inside the wrapped');
        throw 'error';
      })
      .then((value) {
        expect(false, true, reason: 'should complete with an exception if both wrapped and dispose throw');
      }, onError: (error) {
        // Assert
        expect(error, 'dispose error');
      })
      .then(expectAsync1((_) { }));
    });

    test('complete with onError from Future if exception is thrown in dispose but not in wrapped code', () {
      // Arrange
      bool disposed = false;
      IAsyncDisposable disposable = new AsyncDisposable(() => throw 'dispose error');

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false, reason: 'should not have disposed inside the wrapped');
        // nothing...
      })
      .then((value) {
        expect(false, true, reason: 'should complete with an exception');
      }, onError: (error) {
        // Assert
        expect(error, 'dispose error');
      })
      .then(expectAsync1((_) { }));
    });
  });

}